!======================================================================
! Module containing types which hold properties of QCD.
!
! E.g. splitting and coefficient functions.
!
! Details of event shapes
!
module holders
  use types; use consts_dp
  use conv_objects; use event_shapes
  implicit none
  private

  !-- Everything needed to calculate a cross section, given structure
  !   functions and alpha_s.
  !
  !   nomenclature is: objects are leading order, unless there is a suffix n, 
  !   in which case they are NLO.
  type sigma_holder
     type(grid_def) :: gd
     type(PMat) :: P, P1
     type(CMat) :: C2, C2_1, CL_1
     integer    :: factscheme
  end type sigma_holder
  
  !-- for holding everything about a resummed event shape
  !   still in the process of being fleshed out
  type eventshape_holder
     type(CMat) :: C2, C2_1, CL_1
     type(CMat) :: CCA2_1, CCAL_1
     !-- this contains all the various pieces of info that are x-indep
     type(evshp_info) :: es_info
     !-- keep this here even though it is also contained in es_info
     integer    :: id
  end type eventshape_holder

  public :: sigma_holder, holder_InitSigma
  public :: eventshape_holder, holder_InitEventShape

contains
  
  !-------------------------------------------------------
  ! Sets up eveything needed to calculate cross sections
  ! i.e. splitting functions and coefficient functions
  subroutine holder_InitSigma(gd, sh, factscheme)
    use coefficient_functions; use convolution
    use qcd
    type(grid_def),     intent(in)    :: gd
    type(sigma_holder), intent(inout) :: sh
    integer, optional,  intent(in)    :: factscheme
    !-- holds temporary results
    type(grid_conv) :: dconv
    !-- holds all possible combinations of coefficient and splitting functions
    type(grid_conv) :: Cq_Pqq, Cq_Pqg, Cq_Pgq, Cq_Pgg
    type(grid_conv) :: Cg_Pqq, Cg_Pqg, Cg_Pgq, Cg_Pgg

    if (present(factscheme)) then
       sh%factscheme = factscheme
    else
       sh%factscheme = factscheme_default
    end if
    
    sh%gd = gd
    select case (sh%factscheme)
    case (factscheme_MSbar)
       call cobj_InitSplitLO (gd, sh%P)
       call cobj_InitSplitNLO(gd, sh%P1, sh%factscheme)
       call cobj_InitCoeff(gd, sh%C2)
       call cobj_InitCoeff(gd, sh%C2_1, cf_CgF2MSbar, cf_CqF2MSbar)
       call cobj_InitCoeff(gd, sh%CL_1,  cf_CgFL, cf_CqFL)
    case (factscheme_DIS)
       !   The formulae used below comes from notes of 12/06/2000
       !   the principle is that F2 and its derivative must be invariant
       !   on scheme changes (with appropriate modification of the structure
       !   functions).
       !
       !-- first start off with the MSbar functions
       !write(0,'(a)',advance='no') &
       !     &'Constructing DIS scheme NLO splitting functions...'
       call cobj_InitSplitLO (gd, sh%P)
       call cobj_InitSplitNLO(gd, sh%P1, factscheme_MSbar)
       call cobj_InitCoeff(gd, sh%C2)
       call cobj_InitCoeff(gd, sh%C2_1, cf_CgF2MSbar, cf_CqF2MSbar)
       !-- now modify and turn things from MSbar to DIS
       !   collect useful combinations
       !   where possible put the smoother distribution to the right
       !   (only makes a difference when pdfs are non zero at x=1)
       
       call conv_ConvConv(Cq_Pqq, sh%C2_1%q, sh%P%qq)
       call conv_ConvConv(Cq_Pqg, sh%C2_1%q, sh%P%qg)
       call conv_ConvConv(Cq_Pgq, sh%C2_1%q, sh%P%gq)
       call conv_ConvConv(Cq_Pgg, sh%C2_1%q, sh%P%gg)
       call conv_ConvConv(Cg_Pqq, sh%P%qq, sh%C2_1%g)
       call conv_ConvConv(Cg_Pqg, sh%P%qg, sh%C2_1%g)
       call conv_ConvConv(Cg_Pgq, sh%P%gq, sh%C2_1%g)
       call conv_ConvConv(Cg_Pgg, sh%P%gg, sh%C2_1%g)

       !   Pn%qqS -> Pn%qqS + Cg*Pgq + Pqg*Cq
       call conv_AddGridConv(sh%P1%qqS, Cg_Pgq)
       call conv_AddGridConv(sh%P1%qqS, Cq_Pqg)
       !   Pn%qg  -> Pn%qg + Cq*Pqg + Cg*Pgg - Pqq*Cg + 2nf*Pqg*Cg
       call conv_AddGridConv(sh%P1%qg, Cq_Pqg)
       call conv_AddGridConv(sh%P1%qg, Cg_Pgg)
       call conv_AddGridConv(sh%P1%qg, Cg_Pqq, -one)
       call conv_AddGridConv(sh%P1%qg, Cg_Pqg, +2*nf)
       !   Pn%gq  -> Pn%gq - Cq*Pqq - 2*nf*Cg*Pgq-Pgq*Cq + Pgg*Cq
       call conv_AddGridConv(sh%P1%gq, Cq_Pqq, -one)
       call conv_AddGridConv(sh%P1%gq, Cg_Pgq, -2*nf)
       call conv_AddGridConv(sh%P1%gq, Cq_Pgq, -one)
       call conv_AddGridConv(sh%P1%gq, Cq_Pgg, +one)
       !   Pn%gg  -> Pn%gg - (Cg*Pgq + Pqg*Cq)*2nf
       call conv_AddGridConv(sh%P1%gg, Cg_Pgq, -2*nf)
       call conv_AddGridConv(sh%P1%gg, Cq_Pqg, -2*nf)
       !   Pn%qq  -> Pn%qq - beta * Cq
       !   Pn%qg  -> Pn%qg - beta * Cg
       !   Pn%gq  -> Pn%gq + beta * Cq
       !   Pn%gg  -> Pn%gg + beta * Cg * 2nf
       call conv_AddGridConv(sh%P1%qq, sh%C2_1%q, -twopi_beta0)
       call conv_AddGridConv(sh%P1%qg, sh%C2_1%g, -twopi_beta0)
       call conv_AddGridConv(sh%P1%gq, sh%C2_1%q, +twopi_beta0)
       call conv_AddGridConv(sh%P1%gg, sh%C2_1%g, +2*nf*twopi_beta0)

       !   clear unrequired memory
       call conv_DelGridConv(Cq_Pqq)
       call conv_DelGridConv(Cq_Pqg)
       call conv_DelGridConv(Cq_Pgq)
       call conv_DelGridConv(Cq_Pgg)
       call conv_DelGridConv(Cg_Pqq)
       call conv_DelGridConv(Cg_Pqg)
       call conv_DelGridConv(Cg_Pgq)
       call conv_DelGridConv(Cg_Pgg)
       call cobj_DelCoeff(sh%C2_1)

       !write(0,'(a)',advance='yes') ' Done'
       !-- now put in the correct DIS-scheme NLO coefficient functions
       !   fake it as being a zero LO coefficient function...
       call cobj_InitCoeff(sh%C2_1, sh%C2, zero)
       call cobj_InitCoeff(gd, sh%CL_1,  cf_CgFL, cf_CqFL)
    case default
       write(0,*) 'factorisation scheme ',sh%factscheme,&
            &' is not currently supported' 
       stop
    end select
  end subroutine holder_InitSigma
  



  !----------------------------------------------------------------------
  subroutine holder_InitEventShape(esh, sh, id_evshp)
    use event_shapes; use convolution
    type(eventshape_holder), intent(inout) :: esh
    type(sigma_holder), target, intent(in) :: sh
    integer,                 intent(in)    :: id_evshp
    !----------------------------------------------------
    type(grid_def), pointer :: gd

    !-- set the global event shape identifier?
    gd => sh%gd
    call es_setid(id_evshp)
    esh%id = id_evshp
    !-----------------------------------------
    call es_ResummedCoeffs(esh%es_info) 
    call cobj_InitCoeff(gd, esh%C2)  ! just a delta function
    call cobj_InitCoeff(gd, esh%C2_1, es_C2g, es_C2q)
    call cobj_InitCoeff(gd, esh%CL_1, es_CLg, es_CLq)
    !-- sort out the scheme dependence
    if (id_evshp == EvshpId_CEmpty) then
       write(0,*) 'WARNING: CEmpty coefficient function will not have'
       write(0,*) 'the local fact. scheme coefficient function added'
    else
       call cobj_AddCoeff(esh%C2_1, sh%C2_1)
       call cobj_AddCoeff(esh%CL_1, sh%CL_1)
    end if
    
    if (esh%es_info%resumCA) then
       !-- will waste some time setting things to zero, but
       !   within the current framework there is no better way
       !   of doing things. (CA2g, CALg)
       call cobj_InitCoeff(gd, esh%CCA2_1, es_CCA2g, es_CCA2q)
       call cobj_InitCoeff(gd, esh%CCAL_1, es_CCALg, es_CCALq)
       !   will they need something analogous to the scheme dependence
       !   added to them?
    end if
  end subroutine holder_InitEventShape
  
end module holders
