!=======================================================================
! The aim of this routine is to allow easy access to all the information
! that we will need for matching.
!
! This will include a grid_def, a sigma holder, an array of
! eventshape_holders, an array of match_holders
!
! $Id: matching_cache.f90,v 1.4 2001/04/23 13:54:22 salam Exp $
!=======================================================================
module matching_cache
  use types; use consts_dp
  use convolution; use holders; use matching
  implicit none
  public

  integer, parameter :: start_size = 10

  type pdf_item
     real(dp), pointer :: pdf(:,:)
     real(dp)          :: Q
     integer           :: id
  end type pdf_item
  
  !---- various internal caches -----------
  !     (wish we had templates!)
  type pdf_cache
     type(pdf_item), pointer :: items(:)
     integer                 :: nused
  end type pdf_cache
  type esh_cache
     type(eventshape_holder), pointer :: items(:)
     integer                          :: nused
  end type esh_cache

  !-- will need to use this ruse to allow NewCacheItem_mh to return
  !   a pointer to an mh which will persist
  type mh_pointer
     type(match_holder), pointer :: mhp
  end type mh_pointer
  
  type mh_cache
     type(mh_pointer),   pointer :: items(:)
     integer                     :: nused
  end type mh_cache
  !-- for easy reference ---
  interface InitCache
     module procedure InitCache_pdf, InitCache_esh, InitCache_mh
  end interface
  interface NewCacheItem
     module procedure NewCacheItem_pdf, NewCacheItem_esh, NewCacheItem_mh
  end interface
  


  !--- want sh to be a pointer because this allows it to
  !    legitmately referred to in all circumstances (i.e. even if 
  !    match_cache is a copy, the pointer to sh remains valid)
  type match_cache
     type(grid_def)               :: gd
     type(sigma_holder), pointer  :: sh
     type(esh_cache)              :: esh_c
     type(pdf_cache)              :: pdf_c
     type(mh_cache)               :: mh_c
  end type match_cache
  

  public :: match_cache
  public :: mca_InitCache, mca_GetMH

contains
  subroutine mca_InitCache(cache, xmin, dy, factscheme)
    type(match_cache), intent(inout)    :: cache
    real(dp),     intent(in)            :: xmin, dy
    integer,      intent(in), optional  :: factscheme
    !-----------------------------------------------
    integer :: istat

    allocate(cache%sh)

    call conv_InitGridDef(cache%gd, dy = dy, ymax = -log(xmin))
    call holder_InitSigma(cache%gd, cache%sh, factscheme)

    !-- make sure things are clean
    call InitCache(cache%esh_c)
    call InitCache(cache%pdf_c)
    call InitCache(cache%mh_c)
  end subroutine mca_InitCache


  !----------------------------------------------------------------------
  ! Returns pointers to the match_holder and to the pdf distribution
  ! Looks unclean to have have pointer to pdf distribution? But it
  ! saves on having to make multiple calls to this module to get
  ! first one bit then another etc...
  subroutine mca_GetMH(cache, mh, sh, esh, pdf, es_id,pdf_id,xbj,ybj,muF_Q,Q)
    type(match_cache), intent(inout) :: cache
    type(match_holder),   pointer       :: mh
    type(sigma_holder),   pointer       :: sh
    type(eventshape_holder),   pointer  :: esh
    real(dp),             pointer       :: pdf(:,:)
    integer,              intent(in)    :: es_id, pdf_id
    real(dp),             intent(in)    :: xbj, ybj, muF_Q, Q
    !---------------------------------------------------
    type(pdf_item),           pointer :: pdfi
    !type(eventshape_holder),  pointer :: esh
    
    call FindPdf(cache%pdf_c, pdfi, cache%sh, Q*muF_Q, pdf_id)
    call FindEsh(cache%esh_c, esh,  cache%sh, es_id)
    call NewCacheItem(cache%mh_c, mh)
    call ma_InitMatchHolder(mh, cache%sh, esh, pdfi%pdf, xbj, ybj, muF_Q, Q)
    pdf => pdfi%pdf
    sh  => cache%sh
  end subroutine mca_GetMH
  

  

  !======================================================================
  ! Things for locating / or initialising cache items
  ! Leave it up to use to guarantee that a given cache uses consistently
  ! the same grid_def
  subroutine FindPdf(pdf_c, item, sh, Q, pdf_id)
    use parton_distributions
    type(pdf_cache), intent(inout) :: pdf_c
    type(sigma_holder), intent(in) :: sh
    real(dp),        intent(in)    :: Q
    integer,         intent(in)    :: pdf_id
    type(pdf_item),  pointer       :: item
    !-------------------------------------
    integer :: i

    !-- find preexisting version
    do i = 1, pdf_c%nused
       if (pdf_c%items(i)%Q == Q .and. pdf_c%items(i)%id == pdf_id) then
          item => pdf_c%items(i)
          return
       end if
    end do
    !-- else initialise one
    call NewCacheItem(pdf_c, item)
    item%id = pdf_id
    item%Q  = Q
    call pdf_InitDist(sh%gd, item%pdf, Qval = Q, iset = pdf_id)
  end subroutine FindPdf
  !......................................................................
  subroutine FindEsh(esh_c, item, sh, es_id)
    use parton_distributions
    type(esh_cache), intent(inout) :: esh_c
    type(sigma_holder), intent(in) :: sh
    integer,         intent(in)    :: es_id
    type(eventshape_holder),  pointer       :: item
    !-------------------------------------
    integer :: i

    !-- find preexisting version
    do i = 1, esh_c%nused
       if (esh_c%items(i)%id == es_id) then
          item => esh_c%items(i)
          return
       end if
    end do
    !-- else initialise one
    call NewCacheItem(esh_c, item)
    item%id = es_id
    call holder_InitEventShape(item, sh, es_id)
  end subroutine FindEsh
  


  !======================================================================
  ! caching things from here onwards
  !......................................................................
  subroutine InitCache_pdf(cache)
    type(pdf_cache), intent(inout) :: cache
    allocate(cache%items(start_size))
    cache%nused = 0
  end subroutine InitCache_pdf
  subroutine NewCacheItem_pdf(cache, item)
    type(pdf_cache), intent(inout) :: cache
    type(pdf_item),  pointer       :: item
    !---------------------------------------
    integer :: newsize, oldsize
    type(pdf_item),  pointer       :: items(:)
    oldsize = size(cache%items)
    if (cache%nused == oldsize) then
       !-- number is such that when 1st and second are deallocated, third
       !   fits into the space they leave...
       newsize = nint(1.6_dp * oldsize) 
       write(0,*) 'Just called pdf cache amplification'
       allocate(items(newsize))
       items(1:oldsize) = cache%items(:)
       deallocate(cache%items)
       cache%items => items
    end if
    cache%nused =  cache%nused + 1
    item  => cache%items(cache%nused)
  end subroutine NewCacheItem_pdf
  !......................................................................
  subroutine InitCache_esh(cache)
    type(esh_cache), intent(inout) :: cache
    allocate(cache%items(start_size))
    cache%nused = 0
  end subroutine InitCache_esh
  subroutine NewCacheItem_esh(cache, item)
    type(esh_cache), intent(inout) :: cache
    type(eventshape_holder),  pointer       :: item
    !---------------------------------------
    integer :: newsize, oldsize
    type(eventshape_holder),  pointer       :: items(:)
    oldsize = size(cache%items)
    if (cache%nused == oldsize) then
       !-- number is such that when 1st and second are deallocated, third
       !   fits into the space they leave...
       newsize = nint(1.6_dp * oldsize) 
       write(0,*) 'Just called esh cache amplification'
       allocate(items(newsize))
       items(1:oldsize) = cache%items(:)
       deallocate(cache%items)
       cache%items => items
    end if
    cache%nused =  cache%nused + 1
    item  => cache%items(cache%nused)
  end subroutine NewCacheItem_esh
  !......................................................................
  subroutine InitCache_mh(cache)
    type(mh_cache), intent(inout) :: cache
    allocate(cache%items(start_size))
    cache%nused = 0
  end subroutine InitCache_mh
  subroutine NewCacheItem_mh(cache, item)
    type(mh_cache),      intent(inout) :: cache
    type(match_holder),  pointer       :: item
    !---------------------------------------
    integer :: newsize, oldsize
    type(mh_pointer),    pointer       :: items(:)
    oldsize = size(cache%items)
    if (cache%nused == oldsize) then
       !-- number is such that when 1st and second are deallocated, third
       !   fits into the space they leave...
       newsize = nint(1.6_dp * oldsize) 
       write(0,*) 'Just called mh cache amplification'
       allocate(items(newsize))
       items(1:oldsize) = cache%items(:)
       deallocate(cache%items)
       cache%items => items
    end if
    cache%nused =  cache%nused + 1
    allocate(cache%items(cache%nused)%mhp)
    item  => cache%items(cache%nused)%mhp
    !write(0,*) ' cache:',cache%items(cache%nused)%mhp%G22
  end subroutine NewCacheItem_mh

  

end module matching_cache
