!======================================================================
!
! Module for doing everything under the sky with resummed results
!
! $Id: resummed_results.f90,v 1.31 2002/08/01 19:40:02 salam Exp $
!======================================================================

!-- for playing around with a new version of things: sacrifice a little
!   speed in exchange for "a more standard way of doing things"
module resummed_results
  use types; use consts_dp; use qcd
  use matching; use special_functions; use pdf_tabulate
  use event_shapes; use warnings_and_errors; use assertions
  use resummation_definitions
  !-- some variables need special help...
  use brdzE
  implicit none
  private

  
  type RmaxInfo
     real(dp) :: sigma10, sigma20, Rmax, disti1_max, disti2_max
  end type RmaxInfo
  

  public :: rr_R, rr_RMaxInfo, RMaxInfo
  
contains
  
  !-----------------------------------------------------------
  ! Returns the function 
  !               R = (1 +  as C1) Sigma(L,as)
  ! or, including the structure function dependence
  !               R = (q(tau Q^2)/q(Q^2) + as C1) Sigma(L,as)
  !
  ! Does the matching; and the renormalisation scale dependence?
  !
  ! result is normalised to the LO cross section
  !
  ! For some kinds of matching it is necessary to have a certain amount
  ! of extra information. This is provided in RMI, which must be set
  ! up first by calling rr_RmaxInfo
  function rr_R(mh, qtabh, var, alfas, disti1, disti2, &
       &muR_Q, MatchType, RMI) result(res) 
    type(match_holder),       intent(in) :: mh
    type(qtab_holder),        intent(in) :: qtabh
    real(dp),                 intent(in) :: var, alfas
    real(dp),                 intent(in) :: disti1, disti2
    real(dp),                 intent(in) :: muR_Q
    integer,                  intent(in) :: MatchType
    type(RMaxInfo), optional, intent(in) :: RMI
    real(dp)                             :: res
    !--------------------------------------------
    real(dp) :: as2pi, L_basic, L_rscld, lambda
    real(dp) :: expnt1, sigma_q, sigma_q_gamma, sigma_g, log_qscale
    real(dp) :: f2ratio, Qresum, C2L, C2L_gamma
    real(dp) :: sigma1, sigma2, match, disti2_lcl
    real(dp) :: sigmaRg_1, sigmaRg_2, lnres, lnxpd, lnfxd
    integer, save :: warn_id = warn_id_INIT
    integer, save :: warn_id_lambda = warn_id_INIT
    !-- modified matching
    real(dp) :: mod_fact, pp, oldres
    !-- needed for rescaled V
    real(dp) :: lnX, DeltaC1_noP, rescaled_var
    !-- needed for modified lnR matching and some cases of modified M matching.
    real(dp) :: disti2_max_lcl, sigma10, sigma20
    real(dp), parameter :: eps=1e-12_dp

    !-- most basic step in most cases?!
    if (var <= zero) then
       res = zero
       return
    end if

    as2pi   = alfas/twopi

    !--- fixed-order scale dependence ---
    if (muR_Q /= one) then
       disti2_lcl = disti2 + two*twopi_beta0*log(muR_Q) * disti1
    else
       disti2_lcl = disti2
    end if

    !-- fix it up for max values as well...
    if (present(RMI)) then
       disti2_max_lcl = RMI%disti2_max &
            &+ two*twopi_beta0*log(muR_Q) * RMI%disti1_max
    end if

    !--- short cut if just want fixed order ---
    if (MatchType == MatchType_NoResum) then
       res = one + as2pi * (disti1 + as2pi * disti2_lcl)
       !if (var > 0.497_dp) write(0,*) var, res
       return
    end if

    
    !-- How to define our logarithm -----------
    pp = mh%es_info%esPTModPower
    lnX = log(mh%es_info%esRescaleV)
    DeltaC1_noP = mh%es_info%G12*lnX**2+ (mh%es_info%G11)*lnX
    !write(0,*) 'DeltaC1', DeltaC1
    select case (MatchType)
    case(MatchType_Mmod, MatchType_M2mod, MatchType_logRmod)
       L_rscld = log((mh%rescale/(var*mh%es_info%esRescaleV))**pp - &
            &(mh%rescale/(mh%es_info%esmax*mh%es_info%esRescaleV))**pp &
            &   + one) / pp
       L_basic = L_rscld - mh%log_rescale + lnX
    case default
       L_basic = -log(var)
       L_rscld = L_basic + mh%log_rescale - lnX
    end select

    !-- not too sure about wanting to do this, but leave it in for now
    if (L_rscld < -eps) then
       call wae_warn(5,warn_id_lambda,&
            &'rr_R: L_rscld < 0 (==evshp too large);&
            & it has been reset to zero')
       !write(0,*) 'ppp', L_basic, L_rscld, var
       L_rscld = zero
       L_basic = L_rscld - mh%log_rescale + lnX
    end if

    !-- set up some handy resummation shortcuts --
    lambda  = alfas * beta0 * L_rscld
    !if (lambda > half) then
    ! give ourselves some leeroom: we do not trust results that 
    ! are this close to landau pole?
    if (lambda > 0.49_dp) then
       !-- stay to the right of the Landau pole
       res = zero
       return
    end if

    !---- this bit sets up the main properties of the resummed result
    call rr_GetExponents(mh, alfas, L_rscld, sigma_q, sigma_q_gamma, &
         &log_qscale, sigma_g, muR_Q, lnX)

    if (log_qscale == zero) then
       !-- is this really right? Even in improved broadening case?
       !   maybe coefficient of gamma goes to zero in this limit?
       res = (one + as2pi*(mh%C2L_1+DeltaC1_noP-mh%P*lnX))* sigma_q
    else
       Qresum  = mh%Q*mh%muF_Q * exp(log_qscale)
       f2ratio = qtab_Val(qtabh, Qresum) / qtab_Val(qtabh, mh%Q*mh%muF_Q)
       select case (mh%es_info%esC1MuF)
       case(esC1MuF_Q)
          res = (f2ratio + as2pi*(mh%C2L_1+DeltaC1_noP-mh%P*lnX))* sigma_q
          !-- piece requiring extra factor of gamma
          if (sigma_q_gamma /= zero) then
             !res = res + &
             !     &f2ratio * sigma_q_gamma * qtab_Gamma(qtabh, mh%Q*mh%muF_Q)
             res = res + &
                  &f2ratio * sigma_q_gamma * qtab_Gamma(qtabh, Qresum)
          end if
       case(esC1MuF_VQ)
          sigma_q = sigma_q * f2ratio
          ! recall C2L and C2L_gamma are given relative to F2_LO(Qresum)
          call qtab_C2L(qtabh, Qresum, C2L, C2L_gamma)
          res = (one + as2pi*(C2L+DeltaC1_noP)) * sigma_q
          if (lnX*mh%es_info%GP /= zero) res = res &
               & - as2pi*mh%es_info%GP*lnX*qtab_Gamma(qtabh, Qresum)*sigma_q
          if (sigma_q_gamma /= zero) then
             res = res + f2ratio*sigma_q_gamma * &
                  &(qtab_Gamma(qtabh, Qresum) + as2pi * C2L_gamma)
          end if
       end select
    end if
    
    


    !-- following should be relevant only in situation where f2ratio is one
    sigma_g = as2pi*mh%CCA2L_1 * sigma_g
    res     = res + sigma_g

    !-- do matching if requested ------------------
    select case(MatchType)
    case(MatchType_none)
       ! do nothing...
    case(MatchType_M)
       call ma_GetSigmas(mh, L_basic, sigma1, sigma2, muR_Q = muR_Q,  &
            &expnt1   = expnt1, &
            &nlo_evln = (qtabh%qtabid==qtabid_fromPRM .or. &
            &            qtabh%qtabid==qtabid_evlnNLO ))
       match = as2pi*((disti1-sigma1) &
            &         + as2pi*((disti2_lcl-sigma2)-expnt1*(disti1-sigma1)))
       res = res + match * sigma_q
       !if (var > 0.497_dp) write(0,*) var, res
    case(MatchType_Mmod)
       if (.not. present(RMI)) call wae_error('rr_R:', &
            &'for MatchType_Mmod, RMI must be supplied')
       call ma_GetSigmas(mh, L_basic, sigma1, sigma2, muR_Q = muR_Q,  &
            &expnt1   = expnt1, &
            &nlo_evln = (qtabh%qtabid==qtabid_fromPRM .or. &
            &            qtabh%qtabid==qtabid_evlnNLO ))
       mod_fact = one - (var/mh%es_info%esmax)**pp
       !-- the following two things are needed for the broadening...
       match = as2pi*((disti1-sigma1) &
            &         + as2pi*((disti2_lcl-sigma2)&
            &                   -mod_fact*expnt1*(disti1-sigma1)))
       match = match + one + as2pi*(RMI%sigma10 + as2pi*RMI%sigma20) - RMI%Rmax
       res = res + match * sign(abs(sigma_q)**mod_fact,sigma_q)
       !if (var > 0.490_dp) write(0,'(5f12.6)') var, oldres, 1+as2pi*sigma1, &
       !     &   (f2ratio + as2pi*(mh%C2L_1+DeltaC1_noP-mh%P*lnX))* sigma_q
    case(MatchType_M2)
       call ma_GetSigmas(mh, L_basic, sigma1, sigma2, muR_Q = muR_Q,  &
            &expnt1   = expnt1, &
            &nlo_evln = (qtabh%qtabid==qtabid_fromPRM .or. &
            &            qtabh%qtabid==qtabid_evlnNLO ))
       res = res + as2pi*(disti1-sigma1) + &
            &              as2pi**2*(disti2_lcl-sigma2) * sigma_q
    case(MatchType_M2mod)
       if (.not. present(RMI)) call wae_error('rr_R:', &
            &'for MatchType_M2mod, RMI must be supplied')
       call ma_GetSigmas(mh, L_basic, sigma1, sigma2, muR_Q = muR_Q,  &
            &expnt1   = expnt1, &
            &nlo_evln = (qtabh%qtabid==qtabid_fromPRM .or. &
            &            qtabh%qtabid==qtabid_evlnNLO ))
       mod_fact = one - (var/mh%es_info%esmax)**pp
       !-- the following two things are needed for the broadening...
       match = as2pi*(disti1-sigma1) + (&
            &    as2pi**2*(disti2_lcl-sigma2) &
            &    + one + as2pi*(RMI%sigma10 + as2pi*RMI%sigma20) - RMI%Rmax)&
            &    * sigma_q**mod_fact
       res = res + match
    case(MatchType_logR,MatchType_logRmod)
       call ma_GetSigmas(mh, L_basic, sigma1, sigma2, muR_Q = muR_Q,  &
            &expnt1   = expnt1, &
            &sigmaRg_1  = sigmaRg_1, sigmaRg_2 = sigmaRg_2,&
            &nlo_evln = (qtabh%qtabid==qtabid_fromPRM .or. &
            &            qtabh%qtabid==qtabid_evlnNLO ))
       !-- in the limit of this special pathological case, the result
       !   remains what it is
       if (res - sigma_g == zero) return
       !-- arrange this so that it does not crash even if sign has
       !   gone haywire, though of course there is no real meaning to it
       !-- alternative; CCN20-78
       lnres = log(abs(res - sigma_g)) + as2pi*((disti1-sigma1)+&
            &as2pi*((disti2_lcl-sigma2) &
            &       - half*(disti1-sigma1)*(disti1+sigma1-two*sigmaRg_1)))
       !-- fix up normalisation CCN20-90 ---------------------------------
       if (MatchType == MatchType_logRmod) then
          if (.not. present(RMI)) call wae_error('rr_R:', &
               &'for MatchType_logRmod, RMI must be supplied')
          lnres = lnres + &
              &log(1 + (RMI%disti1_max - mh%CCA2L_1)*as2pi &
              &                       + disti2_max_lcl*as2pi**2)&
              & - log(RMI%Rmax - mh%CCA2L_1*as2pi)&
              & - as2pi*(RMI%disti1_max - RMI%sigma10) &
              & - as2pi**2*(disti2_max_lcl - RMI%sigma20 &
              &        - half*(RMI%disti1_max-RMI%sigma10)&
              &                  *(RMI%disti1_max+RMI%sigma10-2*mh%CCA2L_1))
          !-- not quite so perfect matching
       end if
       res = sign(exp(lnres),res - sigma_g) + sigma_g
       !if (var > 0.497_dp) write(0,*) var, res, &
       !     &1 + (RMI%disti1_max - mh%CCA2L_1)*as2pi &
       !       &                       + disti2_max_lcl*as2pi**2+sigma_g
    case(MatchType_diffRsmdExpnd)
       call ma_GetSigmas(mh, L_basic, sigma1, sigma2, muR_Q = muR_Q,  &
            &nlo_evln = (qtabh%qtabid==qtabid_fromPRM .or. &
            &            qtabh%qtabid==qtabid_evlnNLO ))
       res = res - (one + as2pi*(sigma1 + as2pi*sigma2))
    case default
       call wae_error('rr_R: unrecognized matching procedure')
    end select

  end function rr_R
  

  !---------------------------------------------------------------
  ! Function which returns the value of the resummed distribution 
  ! at the point where L_rscld = 0. This information can come in
  ! useful for matching purposes.
  function rr_RMaxInfo(mh, qtabh, alfas, muR_Q, disti1_max, disti2_max&
       &) result(res)
    type(match_holder), intent(in) :: mh
    type(qtab_holder),  intent(in) :: qtabh
    real(dp),           intent(in) :: alfas
    real(dp),           intent(in) :: muR_Q, disti1_max, disti2_max
    type(RmaxInfo)                 :: res
    !-------------------------------------
    real(dp) :: varmax

    !varmax = mh%rescale
    !-- define maximum as point where logs vanish
    varmax = mh%rescale / mh%es_info%esRescaleV
    res%Rmax = rr_R(mh, qtabh, varmax, alfas, disti1=zero, disti2=zero, &
       &muR_Q=muR_Q, MatchType=MatchType_None)
    res%disti1_max = disti1_max
    res%disti2_max = disti2_max
    call ma_GetSigmas(mh, -mh%log_rescale+log(mh%es_info%esRescaleV), &
         &res%sigma10, res%sigma20, muR_Q = muR_Q,  &
         &nlo_evln = (qtabh%qtabid==qtabid_fromPRM .or. &
         &            qtabh%qtabid==qtabid_evlnNLO ))

  end function rr_RMaxInfo
  


  
  !--------------------------------------------------------------
  ! returns the principal information needed for calculating the
  ! resummation
  !
  ! sigma_q multiplies q(qscale*muF_Q)
  ! sigam_q_gamma multiplies gamma * q(qscale*muF_Q)
  subroutine rr_GetExponents(mh, alfas, bigL, sigma_q, sigma_q_gamma, &
       &log_qscale, sigma_g, muR_Q, lnX)
    type(match_holder), intent(in)  :: mh
    real(dp),           intent(in)  :: alfas, bigL
    real(dp),           intent(out) :: sigma_q, sigma_q_gamma, sigma_g 
    real(dp),           intent(out) :: log_qscale
    real(dp),           intent(in)  :: muR_Q, lnX
    !--------------------------------------------
    real(dp) :: R1, R2, Rp, RU1, RU2, RUp, RUD1, RUD2, RUDp
    real(dp) :: Rp_eff, R1bar, R2bar, Rpbar, Lbar, LambdaBar, LambdaBar_gamma
    real(dp) :: ln1ml, ln1m2l
    !-----------------------
    real(dp) :: expnt_q, expnt_g, expnt_q_gamma
    real(dp) :: lambda, sigma_q_sign

    !-- default values -----
    log_qscale = zero
    expnt_g = -1e90_dp
    expnt_q_gamma =  -1e90_dp
    sigma_q_sign = one

    select case (mh%id)
    case(EvshpId_tau_zQ)
       call rr_FullPieces(alfas, bigL, R1, R2, Rp, muR_Q)
       expnt_q = -two*R1 * bigL &
            &    -two*R2 - two*eulergamma*Rp - log(gamma(one+two*Rp))
       expnt_q = expnt_q - lnX * two*Rp
       log_qscale = -half*mh%es_info%GP*bigL
    case(EvshpId_tau_zE)
       call rr_FullPieces(alfas, bigL, R1, R2, Rp, muR_Q)
       call rr_UpperPieces(alfas, bigL, RU1, RU2, RUp, muR_Q)
       expnt_q = -bigL * (R1 + RU1)&
            &  -R2 - RU2 - (eulergamma+ln2)*(Rp+RUp) &
            &    - eulergamma*RUp + log(gamma(one-RUp)/gamma(1+Rp+RUp))
       !-- remember to include the large-angle contribution...
       !   ln1ml and ln1m2l get calculated several times
       ln1ml  = rr_ln1ml(alfas * beta0 * bigL)
       ln1m2l = rr_ln1ml(two*alfas * beta0 * bigL)
       if (mh%es_info%esSumType /= esSumType_NoNonGlobal) then
          expnt_q =  expnt_q + rr_MSLA(half/twopi_beta0 * (ln1ml - ln1m2l))
       end if
       expnt_q = expnt_q - lnX * (Rp+RUp)
       log_qscale = -half*mh%es_info%GP*bigL
    case(EvshpId_tau_tE, EvshpId_Cpr_E, EvshpId_rho_E)
       call rr_FullPieces(alfas, bigL, R1, R2, Rp, muR_Q)
       expnt_q = -bigL * R1&
            &    -R2 - eulergamma*Rp - log(gamma(one+Rp))
       expnt_g = -bigL * R1 * CA/CF
       !-- remember to include the large-angle contribution...
       !   ln1m2l gets calculated several times
       ln1m2l = rr_ln1ml(two*alfas * beta0 * bigL)
       if (mh%es_info%esSumType /= esSumType_NoNonGlobal) then
          expnt_q =  expnt_q + rr_MSLA(-half/twopi_beta0 * (ln1m2l))
       end if
       expnt_q = expnt_q - lnX * (Rp)
    case(EvshpId_brd_zE)
       call rr_UpperPiecesDouble(alfas, bigL, RUD1, RUD2, RUDp, muR_Q)
       select case(mh%es_info%esSumType)
       case(ESSumType_standard)
          expnt_q = -bigL * two*RUD1&
               &    -two*RUD2 - two*RUDp*(eulergamma+half*ln2) &
               &    - lngamma(1+two*RUDp) 
          expnt_q = expnt_q - lnX * (two*RUDp)
          lambda = brdzE_Lambda_Itrplt(two*RUDp)
          sigma_q_sign = sign(one, lambda)
          expnt_q = expnt_q + log(abs(lambda))
          log_qscale = -half*mh%es_info%GP*bigL
       case(ESSumType_WithLbar_m)
          if (lnX /= zero) call wae_error('rr_GetExponents',&
               &'brd_zE, ESSumType_WithLbar_m: esRescaleV must be = 1')
          call brdzE_LambdaEffInfo(alfas, bigL, Lbar, LambdaBar, &
               &LambdaBar_gamma, Rp_eff,&
               &minimal=.true.)
          call rr_UpperPiecesDouble(alfas, Lbar, R1bar, R2bar, Rpbar, muR_Q)
          log_qscale = -half*mh%es_info%GP*Lbar
          expnt_q = -bigL*RUD1 -Lbar*R1bar&
               & -RUD2 - R2bar - RUDp*(eulergamma+ln2) - Rpbar*eulergamma&
               & -lngamma(1+RUDp+Rp_eff)
!!$          !-- tmp for testing
!!$          expnt_q = -bigL*RUD1 -Lbar*R1bar&
!!$               & -RUD2 - R2bar - RUDp*(eulergamma+ln2) - Rp_eff*eulergamma&
!!$               & -lngamma(1+RUDp+Rp_eff)
!!$          LambdaBar_gamma = 1e-300_dp
          !---------------------------------
          expnt_q_gamma = expnt_q + log(LambdaBar_gamma)
          expnt_q = expnt_q + log(LambdaBar)
!!$          expnt_q = -bigL*RUD1 -Lbar*R1bar&
!!$               & -RUD2 - R2bar - RUDp*(eulergamma+ln2) - Rp_eff*eulergamma&
!!$               & -lngamma(1+RUDp+Rp_eff) + log(LambdaBar)
       case(ESSumType_WithLbar)
          if (lnX /= zero) call wae_error('rr_GetExponents',&
               &'brd_zE, ESSumType_WithLbar: esRescaleV must be = 1')
          call brdzE_LambdaEffInfo(alfas, bigL, Lbar, LambdaBar, &
               &LambdaBar_gamma, Rp_eff,&
               &minimal=.false.)
          call rr_UpperPiecesDouble(alfas, Lbar, R1bar, R2bar, Rpbar, muR_Q)
          log_qscale = -half*mh%es_info%GP*Lbar
          !-- safety, though it may lead to some rubbish at low B
          if ((1+RUDp+Rp_eff) <= zero) then
             expnt_q = -300.0_dp
          else
             expnt_q = -bigL*RUD1 -Lbar*R1bar&
              & -RUD2 - R2bar - RUDp*(eulergamma+ln2) - Rpbar*eulergamma&
              & -lngamma(1+RUDp+Rp_eff)
!!$             !-- tmp for testing
!!$             expnt_q = -bigL*RUD1 -Lbar*R1bar&
!!$               & -RUD2 - R2bar - RUDp*(eulergamma+ln2) - Rp_eff*eulergamma&
!!$               & -lngamma(1+RUDp+Rp_eff)
             !-------------------------
             expnt_q_gamma = expnt_q + log(LambdaBar_gamma)
             expnt_q = expnt_q + log(LambdaBar)
!!$             expnt_q = -bigL*RUD1 -Lbar*R1bar&
!!$               & -RUD2 - R2bar - RUDp*(eulergamma+ln2) - Rp_eff*eulergamma&
!!$               & -lngamma(1+RUDp+Rp_eff) + log(LambdaBar)
          end if
       case default
          call wae_error('rr_GetExponents:', &
               &'Unsupported mh%es_info%esSumType for brd_zE')
       end select
    case default
       write(0,*) 'ERROR in rr_R:'
       write(0,*) 'Variable '//es_names(mh%id)//' has not yet been &
            &calculated/implemented'
       stop
    end select

    !-- now we have them in the form we want?
    sigma_q = exp(expnt_q) * sigma_q_sign
    sigma_g = exp(expnt_g)
    sigma_q_gamma = exp(expnt_q_gamma)
  end subroutine rr_GetExponents
  


  !======================================================================
  ! Eqs. C.9 and C.10c from ADS
  subroutine rr_FullPieces(alfas, bigL, R1, R2, Rp, muR_Q)
    use qcd
    real(dp), intent(in)  :: alfas, bigL, muR_Q
    real(dp), intent(out) :: R1, R2, Rp
    real(dp) :: lambda, oneml, onem2l, ln1ml, ln1m2l
    
    lambda = alfas * beta0 * bigL
    if (lambda == zero) then
       R1 = zero; R2 = zero; Rp = zero
       return
    end if
    oneml  = one - lambda
    ln1ml  = rr_ln1ml(lambda)
    onem2l = one - two*lambda
    ln1m2l = rr_ln1ml(two*lambda)

    R1 = cf/(twopi_beta0*lambda) * (onem2l*ln1m2l - two*oneml*ln1ml)
    R2 = 1.5_dp*(cf/twopi_beta0) * ln1ml + &
         & cmw_K*cf*(two*ln1ml - ln1m2l)/(twopi_beta0**2) + &
         & cf*beta1/(twopi*beta0**3) &
         &       * (ln1m2l - two*ln1ml + half*ln1m2l**2 - ln1ml**2)
    Rp = cf/(pi*beta0) * (ln1ml - ln1m2l)
    
    R2 = R2 + lambda*two*log(muR_Q)*(Rp-R1)
  end subroutine rr_FullPieces

  !---------------------------------------------------------------
  ! Eqs. C.7 and C.10a from ADS
  subroutine rr_UpperPieces(alfas, bigL, RU1, RU2, RUp, muR_Q)
    real(dp), intent(in)  :: alfas, bigL, muR_Q
    real(dp), intent(out) :: RU1, RU2, RUp
    real(dp) :: lambda, oneml, ln1ml
    
    lambda = alfas * beta0 * bigL
    if (lambda == zero) then
       RU1 = zero; RU2 = zero; RUp = zero
       return
    end if
    oneml  = one - lambda
    ln1ml  = rr_ln1ml(lambda)

    RU1 = cf/(twopi_beta0*lambda) * (-lambda -ln1ml)
    RU2 = 1.5_dp*(cf/twopi_beta0) * ln1ml +&
         &cmw_K*cf*(lambda + oneml*ln1ml)/(twopi_beta0**2*oneml) + &
         &cf*beta1/(twopi*beta0**3) * &
         &     ( -(lambda + ln1ml)/oneml - half*ln1ml**2)
    RUp = cf/twopi_beta0 * lambda/oneml

    RU2 = RU2 + lambda*two*log(muR_Q)*(RUp-RU1)
  end subroutine rr_UpperPieces

  !---------------------------------------------------------------
  ! same as RU but going "twice as deep"
  subroutine rr_UpperPiecesDouble(alfas, bigL, RUD1, RUD2, RUDp, muR_Q)
    real(dp), intent(in)  :: alfas, bigL, muR_Q
    real(dp), intent(out) :: RUD1, RUD2, RUDp
    real(dp)              :: two_lambda, onem2l, ln1m2l

    two_lambda = two * alfas * beta0 * bigL
    if (two_lambda == zero) then
       RUD1 = zero; RUD2 = zero; RUDp = zero
       return
    end if
    onem2l = one - two_lambda
    ln1m2l = rr_ln1ml(two_lambda)

    RUD1 = two*cf/(twopi_beta0*two_lambda) * (-two_lambda -ln1m2l)
    RUD2 = 1.5_dp*(cf/twopi_beta0) * ln1m2l +&
         &cmw_K*cf*(two_lambda + onem2l*ln1m2l)/(twopi_beta0**2*onem2l) + &
         &cf*beta1/(twopi*beta0**3) * &
         &     ( -(two_lambda + ln1m2l)/onem2l - half*ln1m2l**2)
    RUDp = two*cf/twopi_beta0 * two_lambda/onem2l

    RUD2 = RUD2 + two_lambda*log(muR_Q)*(RUDp-RUD1)
  end subroutine rr_UpperPiecesDouble


  !--------------------------------------------------------------------
  ! calculates ln(1 - lambda)
  ! tries to ensure that one gets a good answer even if lambda is small.x
  function rr_ln1ml(lambda) result(ln1ml)
    real(dp), intent(in) :: lambda
    real(dp) :: ln1ml
    if (abs(lambda) > 1e-2_dp) then
       ln1ml   = log(one - lambda)
    else
       !write(0,*) 'Using it'
       ln1ml  = -lambda - half*lambda**2 - lambda**3/three &
            &- lambda**4/four - lambda**5/5.0_dp - lambda**6/6.0_dp &
            &- lambda**7/7.0_dp
    end if
  end function rr_ln1ml
  

  !---------------------------------------------------------------
  ! multiple soft, large-angle result
  ! First try approximates this by exponentiated double log
  ! Here we just return the log of the result
  ! CCN19-51 and5 7
  function rr_MSLA(as2pi_L) result(res)
    real(dp), intent(in) :: as2pi_L
    real(dp)             :: res
    !--------------------------------------------
    real(dp) :: a2pL_CAover3, a2pL_CA
    ! This must have been a chi-by-eye fit?
    !real(dp), parameter :: a = 1.2_dp, b = 0.3_dp, pow=0.2_dp
    !-- this is what's in the paper...
    real(dp), parameter :: a = 0.85_dp, b = 0.86_dp, pow=1.33_dp

    !as2pi_L = -half/twopi_beta0 * ln1m2l
    !-- calculations were done with CA = 3; if CA is different
    !   then (in large-nc limit) we should be able to set things
    !   right with an apropriate rescaling
    !   Cut it off at zero, to avoid numerical problems
    a2pL_CA      = max(zero,as2pi_L * CA)
    a2pL_CAover3 = a2pL_CA / three


    !-- form taken from SoftDL-resfit. But colour factors
    !   are not put in in a convincing way? NB THIS STILL NEEDS
    !   IMPROVEMENT: FITTED FORM IS NOT PERFECT.
    !res = - pisq/three * CF * CA * as2pi_L**2&
    !     & * (one+a*a2pL_CAover3**2)/(one+b*a2pL_CAover3**pow)
    !write(0,*) as2pi_L*twopi, res/(as2pi_L**2 * pisq/three * CF * CA)

    !-- this is form taken from official parameterisation in NG paper
    res = - pisq/three * CF * CA * as2pi_L**2&
         & * (one+(a*a2pL_CA)**2)/(one+(b*a2pL_CA)**pow)
    
  end function rr_MSLA

end module resummed_results
