!======================================================================
! Consider the following possibility: want special types for
!      set of LO splitting functions 
!      set of coefficient functions

!----------------------------------------------------------------------
! NB: compared to all normal splitting functions we have multiplied 
! by x, to allow for integration measure. NB: this applies also 
! to the plus functions, which do not care about integration measures?
!
! \int dx f(x)_+  g(x) = \int dy [exp(-y)f(exp(-y))]_+ g(exp(-y))
!
!
module splitting_functions
  use types; use consts_dp; use convolution_communicator
  use coefficient_functions; use qcd
  implicit none
  private


  public :: sf_Pgg, sf_Pqq, sf_Pgq, sf_Pqg
  public :: sf_P1qqV, sf_P1gg, sf_P1qqbarV, sf_P1qqS, sf_P1qg, sf_P1gq
  public :: sf_P1qqBryan, sf_P1qgBryan, sf_P1minus
  public :: sf_P1fromg, sf_P1fromq
  public :: sf_P1qqV_DIS, sf_P1qg_DIS

contains
  !----------------------------------------------------------------------
  function sf_Pgg(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    real(dp)             :: x
    x = exp(-y)
    res = zero

    select case(cc_piece)
    case(cc_REAL,cc_REALVIRT)
       res = two*ca*(x/(one-x) + (one-x)/x + x*(1-x))
    end select
    select case(cc_piece)
    case(cc_VIRT,cc_REALVIRT)
       res = res - two*ca*one/(one-x)
    case(cc_DELTA)
       res = (11.0_dp*ca - four*nf*tr)/6.0_dp
    end select

    if (cc_piece /= cc_DELTA) res = res * x
  end function sf_Pgg

  !----------------------------------------------------------------------
  function sf_Pqq(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    real(dp)             :: x
    x = exp(-y)
    res = zero

    select case(cc_piece)
    case(cc_REAL,cc_REALVIRT)
       res = cf*(one+x**2)/(one-x)
    end select
    select case(cc_piece)
    case(cc_VIRT,cc_REALVIRT)
       res = res - cf*two/(one-x)
    case(cc_DELTA)
       res = cf*three*half
    end select

    if (cc_piece /= cc_DELTA) res = res * x
  end function sf_Pqq

  !----------------------------------------------------------------------
  function sf_Pgq(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    real(dp)             :: x
    x = exp(-y)
    res = zero

    select case(cc_piece)
    case(cc_REAL,cc_REALVIRT)
       res = cf*(one + (one-x)**2)/x
    end select
    select case(cc_piece)
    case(cc_VIRT,cc_REALVIRT)
       res = res - zero
    case(cc_DELTA)
       res = zero
    end select

    if (cc_piece /= cc_DELTA) res = res * x
  end function sf_Pgq

  !----------------------------------------------------------------------
  function sf_Pqg(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    real(dp)             :: x
    x = exp(-y)
    res = zero

    select case(cc_piece)
    case(cc_REAL,cc_REALVIRT)
       res = tr*(x**2 + (one-x)**2)
    end select
    select case(cc_piece)
    case(cc_VIRT,cc_REALVIRT)
       res = res - zero
    case(cc_DELTA)
       res = zero
    end select

    if (cc_piece /= cc_DELTA) res = res * x
  end function sf_Pqg
  


  !======================================================================
  ! From here onwards, the NLO splitting functions
  function sf_P1qqV(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    real(dp)             :: x
    real(dp) :: lnx, ln1mx, pqq
    x = exp(-y)
    res = zero

    select case(cc_piece)
    case(cc_REAL,cc_REALVIRT)
       lnx = log(x); ln1mx = log(one - x)
       pqq = two/(one-x) - one - x
!!$       res =    cf**2*( -(two*lnx*ln1mx + 1.5_dp*lnx)*pqq&
!!$            & -(1.5_dp + 3.5_dp*x)*lnx - half*(one+x)*lnx**2 - 5*(1-x))&
!!$            & + cf*ca*( (half*lnx**2 + 11/6.0_dp * lnx&
!!$            &          + 67/18.0_dp - pi**2/6.0_dp) * pqq + (1+x)*lnx&
!!$            &        + 20/three*(1-x))&
!!$            & + cf*tf*(-(two/three*lnx + 10/9.0_dp)*pqq - four/three*(1-x))
res=    CF*Tf*((-1.1111111111111112_dp - (2*lnx)/3._dp)*pqq -          &
     &     (4*(1 - x))/3._dp) +                                        &
     &  CA*CF*((3.7222222222222223_dp + (11*lnx)/6._dp + lnx**2/2._dp -& 
     &        Pi**2/6._dp)*pqq + (20*(1 - x))/3._dp + lnx*(1 + x)) +   &
     &  CF**2*(((-3*lnx)/2._dp - 2*ln1mx*lnx)*pqq - 5*(1 - x) -        &
     &     (lnx**2*(1 + x))/2._dp - lnx*(1.5_dp + (7*x)/2._dp))
    end select
    select case(cc_piece)
    case(cc_VIRT,cc_REALVIRT)
       pqq = -two/(1-x)
       res = res + CA*CF*(3.7222222222222223_dp - Pi**2/6._dp)*pqq - (10*CF&
           & *pqq*Tf)/9._dp
    case(cc_DELTA)
       res = -(CF*(0.16666666666666666_dp + (2*Pi**2)/9._dp)*Tf) + CA&
           & *CF*(0.7083333333333334_dp + (11*Pi**2)/18._dp - 3*zeta3) +&
           & CF**2*(0.375_dp - Pi**2/2._dp + 6*zeta3)
    end select

    if (cc_piece /= cc_DELTA) res = res * x
    
  end function sf_P1qqV


  !======================================================================
  ! DIS version of the above.
  ! Currently the formulae are wrong. Use the complete formulae from 
  ! module holders (where double convolutions are used to work them out).
  function sf_P1qqV_DIS(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    
    write(0,*) 'sf_P1qqV_DIS: DIS scheme splitting functions &
         &currently not supported' 
    stop
    res = sf_P1qqV(y) - cf_CqF2MSbar(y)*(11*CA - 2*nf)/6.0_dp
  end function sf_P1qqV_DIS

  

  !----------------------------------------------------------------------
  function sf_P1gg(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    real(dp)             :: x
    real(dp) :: lnx, ln1mx, pgg, S2x, pggmx
    x = exp(-y)
    res = zero

    select case(cc_piece)
    case(cc_REAL,cc_REALVIRT)
       lnx = log(x); ln1mx = log(one - x)
       pgg   = (one/(one-x) + one/x -two + x*(one-x))
       pggmx = (one/(one+x) - one/x -two - x*(one+x))
       S2x   = sf_S2(x)
       res = CF*Tf*(-16 + 4/(3._dp*x) + 8*x + (20*x**2)/3._dp - lnx**2*(2&
           & + 2*x)- lnx*(6 + 10*x)) + CA*Tf*(2 - (20*pgg)/9._dp - 2*x -&
           & (4*lnx*(1 + x))/3._dp + (26*(-(1/x) + x**2))/9._dp) + CA**2&
           & *(pgg*(7.444444444444445_dp - 4*ln1mx*lnx + lnx**2 - Pi**2&
           & /3._dp) + 2*pggmx*S2x + (27*(1 - x))/2._dp + 4*lnx**2*(1 + x)&
           & + (67*(-(1/x) + x**2))/9._dp - lnx*(8.333333333333334_dp -&
           & (11*x)/3._dp + (44*x**2)/3._dp))
    end select
    select case(cc_piece)
    case(cc_VIRT,cc_REALVIRT)
       pgg = -one/(1-x)
       res = res + CA**2*pgg*(7.444444444444445_dp - Pi**2/3._dp) - (20*CA&
           & *pgg*Tf)/9._dp
    case(cc_DELTA)
       res = (-4*CA*Tf)/3._dp - CF*Tf + CA**2*(2.6666666666666665_dp&
           & + 3*zeta3)
    end select

    if (cc_piece /= cc_DELTA) res = res * x
    
  end function sf_P1gg


  !----------------------------------------------------------------------
  function sf_P1qqbarV(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    real(dp)             :: x
    real(dp) :: lnx, pqqmx, S2x
    x = exp(-y)
    res = zero

    select case(cc_piece)
    case(cc_REAL,cc_REALVIRT)
       lnx = log(x)
       pqqmx = two/(one+x) - one + x
       S2x   = sf_S2(x)
       res = CF*(-CA/2._dp + CF)*(2*pqqmx*S2x + 4*(1 - x) + 2*lnx*(1 + x))
    end select
    select case(cc_piece)
    case(cc_VIRT,cc_REALVIRT)
       res = res + zero
    case(cc_DELTA)
       res = zero
    end select

    if (cc_piece /= cc_DELTA) res = res * x
  end function sf_P1qqbarV

  !----------------------------------------------------------------------
  function sf_P1qqS(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    real(dp)             :: x
    real(dp) :: lnx, ln1mx
    x = exp(-y)
    res = zero

    select case(cc_piece)
    case(cc_REAL,cc_REALVIRT)
       lnx = log(x); ln1mx = log(one - x)
       res = (CF*TR*(20 - 9*(2 - lnx + lnx**2)*x - 9*(-6 - 5*lnx + lnx&
           & **2)*x**2 + 8*(-7 + 3*lnx)*x**3))/(9._dp*x) 
    end select
    select case(cc_piece)
    case(cc_VIRT,cc_REALVIRT)
       res = res + zero
    case(cc_DELTA)
       res = zero
    end select

    if (cc_piece /= cc_DELTA) res = res * x
  end function sf_P1qqS


  !----------------------------------------------------------------------
  function sf_P1qg(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    real(dp)             :: x
    real(dp) :: lnx, ln1mx, pqg, pqgmx, S2x
    x = exp(-y)
    res = zero

    select case(cc_piece)
    case(cc_REAL,cc_REALVIRT)
       lnx = log(x); ln1mx = log(one - x)
       pqg   = x**2 + (one-x)**2
       pqgmx = x**2 + (one+x)**2
       S2x   = sf_S2(x)
       res = CF*(half*TR)*(4 + 4*ln1mx + (10 - 4*(ln1mx - lnx) + 2*(-ln1mx&
            & + lnx)& 
           & **2 - (2*Pi**2)/3._dp)* pqg - lnx*(1 - 4*x) - lnx**2*(1 - 2&
           & *x) - 9*x) + CA*(half*TR)*(20.22222222222222_dp - 4*ln1mx + (&
           & -24.22222222222222_dp + 4*ln1mx - 2*ln1mx**2 + (44*lnx)/3._dp&
           & - lnx**2 + Pi**2/3._dp)*pqg + 2*pqgmx*S2x + 40/(9._dp*x) +&
           & (14*x)/9._dp - lnx**2*(2 + 8*x) + lnx*(-12.666666666666666_dp&
           & + (136*x)/3._dp))
    end select
    select case(cc_piece)
    case(cc_VIRT,cc_REALVIRT)
       res = res + zero
    case(cc_DELTA)
       res = zero
    end select

    if (cc_piece /= cc_DELTA) res = res * x
  end function sf_P1qg


  !======================================================================
  ! DIS version of the above
  function sf_P1qg_DIS(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    
    write(0,*) 'sf_P1qg_DIS: DIS scheme splitting functions currently not supported' 
    stop
    res = sf_P1qg(y) - cf_CgF2MSbar(y)*(11*CA - 2*nf)/6.0_dp
  end function sf_P1qg_DIS


  !----------------------------------------------------------------------
  function sf_P1gq(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    real(dp)             :: x
    real(dp) :: lnx, ln1mx, pgq, pgqmx,S2x
    x = exp(-y)
    res = zero

    select case(cc_piece)
    case(cc_REAL,cc_REALVIRT)
       lnx = log(x); ln1mx = log(one - x)
       pgq   = (one + (one-x)**2)/x
       pgqmx = -(one + (one+x)**2)/x
       S2x   = sf_S2(x)
       res = CF*Tf*(-((2.2222222222222223_dp + (4*ln1mx)/3._dp)*pgq) - (4&
           & *x)/3._dp) + CF**2*(-2.5_dp - (3*ln1mx + ln1mx**2)*pgq - lnx&
           & **2*(1 - x/2._dp) - (7*x)/2._dp - 2*ln1mx*x + lnx*(2 + (7*x)&
           & /2._dp)) + CA*CF*(3.111111111111111_dp + pgq*(0.5_dp + (11&
           & *ln1mx)/3._dp + ln1mx**2 - 2*ln1mx*lnx + lnx**2/2._dp - Pi**2&
           & /6._dp) + pgqmx*S2x + (65*x)/18._dp + 2*ln1mx*x + (44*x**2)&
           & /9._dp + lnx**2*(4 + x) - lnx*(12 + 5*x + (8*x**2)/3._dp))
    end select
    select case(cc_piece)
    case(cc_VIRT,cc_REALVIRT)
       res = res + zero
    case(cc_DELTA)
       res = zero
    end select

    if (cc_piece /= cc_DELTA) res = res * x
  end function sf_P1gq


  !======================================================================
  ! Alternative versions of the splitting functions, more similar
  ! to what is in the ESW book
  ! and some things for checking sum rules more easily
  function sf_P1qqBryan(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    res = sf_P1qqV(y) + sf_P1qqbarV(y) + two*nf * sf_P1qqS(y)
    !res = sf_P1qqBryanTyped(y)
  end function sf_P1qqBryan
  function sf_P1qgBryan(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    res = two*nf * sf_P1qg(y)
  end function sf_P1qgBryan
  function sf_P1minus(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    res = sf_P1qqV(y) - sf_P1qqbarV(y)
  end function sf_P1minus
  function sf_P1fromq(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    res = sf_P1qqBryan(y) + sf_P1gq(y)
  end function sf_P1fromq
  function sf_P1fromg(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    res = sf_P1gg(y) + sf_P1qgBryan(y)
  end function sf_P1fromg


  !----------------------------------------------------------------------
  ! Following was useful for debugging purposes...
  function sf_P1qqBryanTyped(y) result(res)
    real(dp), intent(in) :: y
    real(dp)             :: res
    real(dp)             :: x
    real(dp) :: lnx, ln1mx, pqq, pqqmx, S2x
    x = exp(-y)
    res = zero

    select case(cc_piece)
    case(cc_REAL,cc_REALVIRT)
       lnx = log(x); ln1mx = log(one - x)
       pqq = two/(one-x) - one - x
       pqqmx = two/(one+x) - one + x
       S2x = sf_S2(x)
       res = CA*CF*((3.7222222222222223_dp + (11*lnx)/6._dp + lnx**2/2._dp&
            & - Pi**2/6._dp)*pqq - pqqmx*S2x + (14*(1 - x))/3._dp) +&
            &  CF**2*(-1 - ((3*lnx)/2._dp + 2*ln1mx*lnx)*pqq + 2*pqqmx*S2x + &
            &     lnx*(0.5_dp - (3*x)/2._dp) + x - (lnx**2*(1 + x))/2._dp) +&
            & CF*Tf*(-5.333333333333333_dp - &
            &     (1.1111111111111112_dp + (2*lnx)/3._dp)*pqq + 40/(9._dp*x) +&
            &     (40*x)/3._dp - (112*x**2)/9._dp - 2*lnx**2*(1 + x) +&
            &     lnx*(2 + 10*x + (16*x**2)/3._dp))

    end select
    select case(cc_piece)
    case(cc_VIRT,cc_REALVIRT)
       pqq = -two/(1-x)
       res = res + CA*CF*(3.7222222222222223_dp - Pi**2/6._dp)*pqq - (10*CF&
           & *pqq*Tf)/9._dp
    case(cc_DELTA)
       res = -(CF*(0.16666666666666666_dp + (2*Pi**2)/9._dp)*Tf) + CA&
           & *CF*(0.7083333333333334_dp + (11*Pi**2)/18._dp - 3*zeta3) +&
           & CF**2*(0.375_dp - Pi**2/2._dp + 6*zeta3)
    end select

    if (cc_piece /= cc_DELTA) res = res * x
    
  end function sf_P1qqBryanTyped

  

  !======================================================================
  ! helper...
  function sf_S2(x) result(S2)
    use special_functions
    real(dp), intent(in) :: x
    real(dp)             :: S2, lnx
    lnx = log(x)
    S2 = -two*ddilog(-x) + half * lnx**2 - two*lnx*log(one+x) - pi**2&
         & /6.0_dp 
  end function sf_S2
  
end module splitting_functions




